#pragma once

#include <Arduino.h>

typedef void (*ActionCb)(void *data);

class Button {
public:
    //! Create a Button with an INPUT_PULLUP pin associated.Select if button is active LOW (default) or HIGH. An optional callback method can be configured (s.a. setAction()).
    Button(byte pin, byte activeLevel = LOW, ActionCb a = NULL)
        : mPin(pin)
        , mActiveLevel(activeLevel)
        , mAction(a)
        , mState(0)
        , mDebounceDelay(50)  // 50 ms
        , mDebounceTimeout(0)
    {
        pinMode(mPin, INPUT_PULLUP);
    }

    //! Needs to by called cyclically to detect a state change and signal debuoncing. Returns true once, if the button is pressed.
    bool checkIfPressed()
    {
        bool checkResult = false;

        if (mDebounceTimeout == 0) {
            // system will detect if button is pressed
            byte buttonState = (digitalRead(mPin) == mActiveLevel) ? HIGH : LOW;            
            if (mState != buttonState) {
                mDebounceTimeout = millis() + mDebounceDelay;
                mState = buttonState;
                // trigger action on rising edge
                if (mState == HIGH) {
                    checkResult = true;
                    if (mAction != NULL) (*mAction)(mActionData);
                }  
            }
        }
        else {
            // during debouncing period ignore any button signal change
            if (millis() > mDebounceTimeout) {
                mDebounceTimeout = 0;
            }
        }

        return checkResult;
    }

    //! Returns the current (debounced) state of the button (true if pressed)
    bool isPressed() {
        return (mState == HIGH);
    }

    //! Registers a callback mathod that is triggered if button is pressed. Set to NULL to disable.
    void setAction(ActionCb a, void *actionData) {
        mAction = a;
        mActionData = actionData;
    }

private:
    byte mPin;              
    byte mActiveLevel;                  //!< (e.g. LOW if the button is low-active)
    ActionCb mAction;
    void *mActionData;
    byte mState;                        //!< current debounced button state (remains HIGH as long as button is pressed)
    unsigned int mDebounceDelay;
    unsigned long mDebounceTimeout;     //!< Controls the debouncing. 0: no debouncing is active; != 0: time until debouncing is active
    Button(const Button& c);
    Button& operator=(const Button& c);
};
